// inetPort.cpp

/*************************************************************************
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory, and the Regents of the University of California, as
* Operator of Los Alamos National Laboratory. MPFserial is distributed
* subject to a Software License Agreement found in file LICENSE that
* is included with this distribution.
*************************************************************************/
/*
    Author: Marty Kraimer
    Date: 23JAN2003

*/

#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>

#include <iocsh.h>
#include <epicsThread.h>
#include <epicsEvent.h>
#include <epicsMutex.h>
#include <epicsTime.h>
#include <epicsAssert.h>
#include <iocsh.h>
#include <mpfType.h>
#include <Tcp.h>
#include "SerialPort.h"

extern "C" { 
int initInetPort(char *portName, char *inetAddr,int port,int bufsize);
}

class InetPort : public SerialPort{
public:
    InetPort(char *portName,char *inetAddr,int port,
        Tcp *pTcp,int bufsize);
    virtual serialStatus write(
	unsigned char *data, int nbytes, double timeout);
    virtual serialStatus read(double timeout);
    virtual void flush();
    char *portName;
    char *inetAddr;
    int port;
    Tcp *pTcp;
    tcpConnectState inetState;
    int bufsize;
    int nextFree;
    char *buffer;
    epicsEventId dataArrived;
    epicsMutexId lock;
};

int initInetPort(char *portn,char *inet,int port,int bufsize)
{
    char *portName,*inetAddr;
    portName = (char *)calloc(strlen(portn)+1,sizeof(char));
    strcpy(portName,portn);
    inetAddr = (char *)calloc(strlen(inet)+1,sizeof(char));
    strcpy(inetAddr,inet);
    Tcp *pTcp = new Tcp(false,port,inetAddr,bufsize);
    InetPort *pInetPort = new InetPort(portName,inetAddr,port,pTcp,bufsize);
    pInetPort->registerName(portName);
    return(0);
}

static void connectCallback(void *pvt,tcpConnectState state)
{
    InetPort *pInetPort = (InetPort *)pvt;
    epicsMutexMustLock(pInetPort->lock);
    pInetPort->inetState = state;
    epicsMutexUnlock(pInetPort->lock);
}

static void receiveCallback(void *pvt,void *buffer,int32 nbytes)
{
    InetPort *pInetPort = (InetPort *)pvt;
    epicsMutexMustLock(pInetPort->lock);
    if(nbytes > (pInetPort->bufsize - pInetPort->nextFree)) {
        printf("InetPort %s %s %d : received %d but bufsize %d nextFree %d\n",
            pInetPort->portName,pInetPort->inetAddr,pInetPort->port,
            nbytes,pInetPort->bufsize,pInetPort->nextFree);
        nbytes = pInetPort->bufsize - pInetPort->nextFree;
    }
    memcpy(pInetPort->buffer +pInetPort->nextFree,buffer,nbytes);
    pInetPort->nextFree += nbytes;
    epicsMutexUnlock(pInetPort->lock);
    epicsEventSignal(pInetPort->dataArrived);
}

InetPort::InetPort(char *port,char *inet, int p,
    Tcp *ptcp,int size)
: portName(port), inetAddr(inet), port(p),
  pTcp(ptcp), inetState(tcpConnectNo), bufsize(size), nextFree(0)
{
    buffer = new char[bufsize];
    dataArrived = epicsEventMustCreate(epicsEventEmpty);
    lock = epicsMutexMustCreate();
    pTcp->connect(this,connectCallback,receiveCallback);
}

serialStatus InetPort::write(
    unsigned char *data, int nbytes, double )
{
    if(inetState!=tcpConnectYes)  return(serialFailure);
    int status = pTcp->sendFlush(data,nbytes);
    if(status) return(serialFailure);
    if(sniffHandler) {
        for(int ind=0; ind<nbytes; ind++) {
            (*sniffHandler)(sniffHandlerParm,false,data[ind]);
        }
    }
    return(serialSuccess);
}

serialStatus InetPort::read(double timeOut)
{
    byteHandlerRC rc = byteHandlerError;
    double timeLeft = timeOut;
    epicsTimeStamp start,now;

    assert(byteHandler);
    if(inetState!=tcpConnectYes)  return(serialFailure);
    epicsTimeGetCurrent(&start);
    while(1) {
        epicsEventWaitStatus status;
        status = epicsEventWaitWithTimeout(dataArrived,timeLeft);
        if(status!=epicsEventWaitOK) {
            return((status==epicsEventWaitTimeout)
                   ? serialTimeout : serialFailure);
        }
        epicsMutexMustLock(lock);
        for(int ind=0; ind<nextFree; ind++) {
            if(sniffHandler) (*sniffHandler)
                (sniffHandlerParm,true,(unsigned char)buffer[ind]);
            rc = (*byteHandler)(byteHandlerParm,(unsigned char)buffer[ind]);
            if(rc!=byteHandlerOK) {
                if(rc==byteHandlerEndRead) {
                    int nextNotUsed = ind + 1;
                    if(nextNotUsed<nextFree) {
                        int nleft = nextFree - nextNotUsed + 1;
                        memmove(buffer,&buffer[nextNotUsed],nleft);
                        nextFree = nleft;
                    } else {
                        nextFree = 0;
                    }
                }
                serialStatus ioStatus = (rc==byteHandlerEndRead)
                    ? serialSuccess : serialFailure;
                epicsMutexUnlock(lock);
                return(ioStatus);
            }
        }
        nextFree = 0;
        epicsMutexUnlock(lock);
        epicsTimeGetCurrent(&now);
        timeLeft = epicsTimeDiffInSeconds(&now,&start);
    }
}

void InetPort::flush()
{
    nextFree = 0; pTcp->flush();
}

extern "C" {
static const iocshArg initInetPortArg0 = { "portName",iocshArgString};
static const iocshArg initInetPortArg1 = { "inetAddr",iocshArgString};
static const iocshArg initInetPortArg2 = { "port",iocshArgInt};
static const iocshArg initInetPortArg3 = { "bufsize",iocshArgInt};
static const iocshArg * const initInetPortArgs[4] = {
    &initInetPortArg0,&initInetPortArg1,&initInetPortArg2,
    &initInetPortArg3};
static const iocshFuncDef initInetPortFuncDef = {
    "initInetPort",4,initInetPortArgs};
static void initInetPortCallFunc(const iocshArgBuf *args)
{
    initInetPort(args[0].sval,args[1].sval,args[2].ival,args[3].ival);
}

static void inetPortRegister(void)
{
    iocshRegister(&initInetPortFuncDef,initInetPortCallFunc);
}
} // end extern "C"

class MpfSerialRegister {
public:
    MpfSerialRegister() {inetPortRegister();}
};

static MpfSerialRegister inetPortRegisterObj;

