// serialServer.cpp

/*************************************************************************
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory, and the Regents of the University of California, as
* Operator of Los Alamos National Laboratory. MPFserial is distributed
* subject to a Software License Agreement found in file LICENSE that
* is included with this distribution.
*************************************************************************/
// Author: Marty Kraimer 
// Date: 7/98


#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>

#include <epicsInterrupt.h>
#include <epicsThread.h>
#include <epicsTime.h>
#include <epicsRingBytes.h>
#include <epicsAssert.h>
 
#include "Message.h"
#include "Int32Message.h"
#include "Char8ArrayMessage.h"
#include "SerialConfigMessage.h"
#include "SerialPort.h"
#include "serialServer.h"
 
class Serial;

extern "C" {
static void serialServerThread(Serial *pSerial);
}

class Serial
{
public:
    Serial(const char *name,const char *portName,
	int bufsize, int queueSize, const char *eomstr);
    void processConfig(SerialConfigMessage *);
    void processGetConfig(Int32Message *);
    void processChar8Array(Char8ArrayMessage *);
    bool getStartOK() const {return(startOK);}
private:
    const char *serverName;
    static byteHandlerRC byteHandler(void*,unsigned char);
    MessageServer *pMessageServer;
    SerialPort* pSerialPort;
    bool startOK;
    bool gotFirstEom;
    serialStatus status;
    unsigned char eomString[2];
    int eomLen;
    int exactSizeToRead;
    int nextIn;
    int readBufSize;
    unsigned char *readBuf;
    friend void serialServerThread(Serial *pSerial);
};

static char taskname[] = "serial";
extern "C" int initSerialServer(const char *serverName,const char *portName,
    int bufsize, int queueSize, const char *eomstr)
{
    if(bufsize<=0) {
        printf("%s bufsize must be > 0\n",serverName);
        return(0);
    }
    Serial *pSerial =
	new Serial(serverName,portName,bufsize,queueSize, eomstr);
    if(!pSerial->getStartOK()) return(0);
    epicsThreadId threadId = epicsThreadCreate(
        taskname,epicsThreadPriorityMedium,
        epicsThreadGetStackSize(epicsThreadStackMedium),
        (EPICSTHREADFUNC)serialServerThread,(void *)pSerial);
    if(!threadId)
	printf("%s serialServer epicsThreadCreate Failure\n",serverName);
    return(0);
}

Serial:: Serial(const char *name,const char *portName,
    int bufsize, int queueSize, const char *eomstr)
:   serverName(name), startOK(false), gotFirstEom(false),
    status(serialSuccess), exactSizeToRead(0),
    nextIn(0), readBufSize(0), readBuf(0)
{
    if(eomstr==0) {
	eomLen = 0;
    } else switch(strlen((const char  *)eomstr)) {
    case 0:
        eomString[0] = 0; eomLen = 1; break;
    case 1:
        eomString[0] = eomstr[0]; eomLen = 1; break;
    case 2:
        eomString[0] = eomstr[0]; eomString[1] = eomstr[1]; eomLen = 2; break;
    default:
	printf("%s serialServer illegal length eomstr. eomLen set to 0\n",
	    serverName);
	eomString[0] = 0; eomLen = 0; break;
    }
    readBuf = (unsigned char *)(new char[(size_t)bufsize]);
    readBufSize = bufsize;
    ::memset((void *)readBuf,'\0',(size_t)readBufSize);
    pSerialPort = SerialPort::bind(portName,this,byteHandler);
    if(pSerialPort==0) {
	printf("%s serialServer: could not bind to SerialPort\n",serverName);
        return;
    }
    pMessageServer = new MessageServer(name,queueSize);
    startOK = true;
}

void serialServerThread(Serial *pSerial)
{
    while(true) {
        pSerial->pMessageServer->waitForMessage();
	Message *message;
        while((message = pSerial->pMessageServer->receive())) {
            switch(message->getType()) {
	    case messageTypeSerialConfig:
	        pSerial->processConfig((SerialConfigMessage *)message);
	        break;
	    case messageTypeInt32:
	        pSerial->processGetConfig((Int32Message *)message);
	        break;
	    case messageTypeChar8Array:
	        pSerial->processChar8Array((Char8ArrayMessage *)message);
	        break;
            default:
                printf("%s serialServer got illegal message type %d\n",
                    pSerial->serverName, message->getType());
                break;
	    }
	    delete message;
	}
    }
}

void Serial::processConfig(SerialConfigMessage *pconfig)
{
    Int32Message *preply = (Int32Message *)pMessageServer->
	allocReplyMessage(pconfig,messageTypeInt32);
    bool result = pSerialPort->config(
	    (int)pconfig->baud,(int)pconfig->stopBits,
	    (int)pconfig->bitsPerChar,(char)pconfig->parity,
            (char)pconfig->flowControl);
    preply->status = (result ? 0 : 1);
    pMessageServer->reply(preply);

}
void Serial::processGetConfig(Int32Message *pconfig)
{
    int baud = 0;
    int stopBits = 0;
    int bitsPerChar = 0;
    char parity = 0;
    char flowControl = 0;
    SerialConfigMessage *preply = (SerialConfigMessage *)pMessageServer->
	allocReplyMessage(pconfig,messageTypeSerialConfig);
    pSerialPort->getConfig(&baud,&stopBits,&bitsPerChar,&parity,&flowControl);
    preply->baud = baud;
    preply->stopBits = stopBits;
    preply->bitsPerChar = bitsPerChar;
    preply->parity = parity;
    preply->flowControl = flowControl;
    pMessageServer->reply(preply);
}

byteHandlerRC Serial::byteHandler(void* v, unsigned char data)
{
    Serial *pSerial = (Serial *)v;
    int nextIn = pSerial->nextIn;
    if(nextIn>=pSerial->readBufSize) {
        pSerial->status = serialOverflow;
        return(byteHandlerEndRead);
    }
    pSerial->readBuf[nextIn] = data;
    ++pSerial->nextIn;
    if(pSerial->eomLen>0) {
        if(!pSerial->gotFirstEom) {
            if(data==pSerial->eomString[0]) {
                if(pSerial->eomLen==1)  goto goteom;
                pSerial->gotFirstEom = true;
            }
        } else {
            if(data==pSerial->eomString[1]) goto goteom;
            pSerial->gotFirstEom = false;
        }
    }
    if((pSerial->exactSizeToRead>0)
    && (pSerial->nextIn >= pSerial->exactSizeToRead) )goto goteom;
    return(byteHandlerOK);
goteom:
    pSerial->gotFirstEom = false;
    return(byteHandlerEndRead);
}

void Serial::processChar8Array(Char8ArrayMessage *preceive)
{
//MARTY CHECK status values carefully
    int cmd = preceive->cmd;
    serialStatus portStatus = serialSuccess;
    Char8ArrayMessage *preply = (Char8ArrayMessage *)pMessageServer->
        allocReplyMessage(preceive,messageTypeChar8Array);
    if(cmd&cmdSetEom) {
        eomLen = preceive->eomLen;
        eomString[0] = preceive->eomString[0];
        eomString[1] = preceive->eomString[1];
    }
    if(cmd&cmdFlush) pSerialPort->flush();
    if(cmd&cmdWrite) {
        if(preceive->timeoutUnits==timeoutUnitsMilliseconds) {
            portStatus = pSerialPort->write((unsigned char *)preceive->value,
         preceive->getSize(), ((double)(preceive->timeout/1000.0)));
        } else {
            portStatus = pSerialPort->write((unsigned char *)preceive->value,
         preceive->getSize(), (int)preceive->timeout);
        }
    }
    if(portStatus!=serialSuccess || !(cmd&cmdRead)) {
        preply->status = portStatus;
        pMessageServer->reply(preply);
        return;
    }
    assert(cmd&cmdRead);
    exactSizeToRead = preceive->extra;
    nextIn = 0;
    status = serialSuccess;
    if(preceive->timeoutUnits==timeoutUnitsMilliseconds) {
        portStatus = pSerialPort->read(((double)preceive->timeout/1000.0));
    } else {
        portStatus = pSerialPort->read((int)preceive->timeout);
    }
    if(portStatus!=serialSuccess) {
        preply->status = portStatus;
        pMessageServer->reply(preply);
        return;
    }
    preply->status = status;
    int len = nextIn;
    preply->allocValue(len);
    ::memcpy((void *)preply->value,(void *)readBuf,len);
    preply->setSize(len);
    pMessageServer->reply(preply);
    return;
}

#include <iocsh.h>
extern "C" {

static const iocshArg initSerialServerArg0 = { "serverName",iocshArgString};
static const iocshArg initSerialServerArg1 = { "portname",iocshArgString};
static const iocshArg initSerialServerArg2 = { "bufSize",iocshArgInt};
static const iocshArg initSerialServerArg3 = { "queueSize",iocshArgInt};
static const iocshArg initSerialServerArg4 = { "eomString",iocshArgString};
static const iocshArg * const initSerialServerArgs[5] = {
    &initSerialServerArg0,&initSerialServerArg1,&initSerialServerArg2,
    &initSerialServerArg3,&initSerialServerArg4};
static const iocshFuncDef initSerialServerFuncDef = {
    "initSerialServer",5,initSerialServerArgs};
static void initSerialServerCallFunc(const iocshArgBuf *args)
{
    initSerialServer(
        args[0].sval,args[1].sval,args[2].ival,
        args[3].ival,args[4].sval);
}

static void serialServerRegister(void)
{
    iocshRegister(&initSerialServerFuncDef,initSerialServerCallFunc);
}
} // extern "C"

class serverRegister {
public:
    serverRegister() {serialServerRegister();}
};

static serverRegister serialServerRegisterObj;

